# -*- coding: utf-8 -*-

"""
iquery.utils
~~~~~~~~~~~~~

A simple args parser and a color wrapper.
"""

import sys


__all__ = ['args', 'colored', 'exit_after_echo']


def exit_after_echo(msg, color='red'):
    if color == 'red':
        print(colored.red(msg))
    else:
        print(msg)
    exit()


class Args(object):

    """A simple customed args parser for `iquery`."""

    def __init__(self, args=None):
        self._args = sys.argv[1:]

    def __repr__(self):
        return '<args {}>'.format(repr(self._args))

    def __len__(self):
        return len(self._args)

    @property
    def all(self):
        return self._args

    def get(self, idx):
        try:
            return self.all[idx]
        except IndexError:
            return None

    @property
    def is_null(self):
        return len(self.all) == 0

    @property
    def options(self):
        """Train tickets query options."""
        arg = self.get(0)
        if arg.startswith('-') and not self.is_asking_for_help:
            return arg[1:]
        return ''.join(x for x in arg if x in 'dgktz')

    def contain_show_type(self):
        arg = self.get(2)
        if is_show_type(arg):
            return arg
        return None

    @property
    def is_asking_for_help(self):
        arg = self.get(0)
        if arg in ('-h', '--help'):
            return True
        return False

    @property
    def is_querying_show(self):
        from .showes import is_show_type
        arg = self.get(1)
        if len(self) not in (2, 3):
            return False
        if is_show_type(arg):
            return True
        return False

    @property
    def is_querying_train(self):
        l = len(self)
        if l not in (3, 4):
            return False
        if self.is_querying_show:
            return False
        if l == 4:
            arg = self.get(0)
            if not arg.startswith('-'):
                return False
            if arg[1] not in 'dgktz':
                return False
        return True

    @property
    def is_querying_movie(self):
        arg = self.get(0)
        if arg in ('-m', '电影'):
            return True
        return False

    @property
    def is_querying_putian_hospital(self):
        return self.get(0) == '-p' and len(self) in (2, 3)

    @property
    def as_train_query_params(self):
        opts = self.options
        if opts:
            # apped valid options to end of list
            return self._args[1:] + [opts]
        return self._args

    @property
    def as_show_query_params(self):
        return self._args

    @property
    def as_hospital_query_params(self):
        return self._args[1:]


class Colored(object):

    """Keep it simple, only use `red` and `green` color."""

    RED = '\033[91m'
    GREEN = '\033[92m'

    #: no color
    RESET = '\033[0m'

    def color_str(self, color, s):
        return '{}{}{}'.format(
            getattr(self, color),
            s,
            self.RESET
        )

    def red(self, s):
        return self.color_str('RED', s)

    def green(self, s):
        return self.color_str('GREEN', s)


args = Args()
colored = Colored()



# -*- coding: utf-8 -*-

"""
iquery.showes
~~~~~~~~~~~~~~

Show tickets query and display. The datas come
from:
    www.damain.cn
"""

import os
import re
import sys
import requests

from .utils import colored, exit_after_echo
from bs4 import BeautifulSoup
from prettytable import PrettyTable
from datetime import datetime, timedelta
from requests.exceptions import ConnectionError


__all__ = ['is_show_type', 'query']


SHOWES_QUERY_URL = 'http://www.damai.cn/projectlist.do'

# ERROR MSG
QUERY_DAYS_INVALID = 'Invalid days.'
CITY_NOT_FOUND = 'Sorry, your city is not supported.'
SHOW_NOT_FOUND = 'No result.'
NETWORK_CONNECTION_FAIL = 'Network connection failed.'

# All supported show types and its query params.
SHOW_TYPES = {
    '演唱会': {'mcid': 1, 'ccid': ''},
    '音乐会': {'mcid': 2, 'ccid': ''},
    '音乐剧': {'mcid': 3, 'ccid': 22},
    '儿童剧': {'mcid': 3, 'ccid': 23},
    '歌舞剧': {'mcid': 3, 'ccid': 21},
    '话剧': {'mcid': 3, 'ccid': 19},
    '歌剧': {'mcid': 3, 'ccid': 20},
    '舞蹈': {'mcid': 4, 'ccid': ''},
    '相声': {'mcid': 5, 'ccid': 27},
    '魔术': {'mcid': 5, 'ccid': 28},
    '马戏': {'mcid': 5, 'ccid': 29},
    '杂技': {'mcid': 5, 'ccid': 30},
    '戏曲': {'mcid': 5, 'ccid': 31},
    '比赛': {'mcid': 6, 'ccid': ''}
}


is_show_type = frozenset(SHOW_TYPES.keys()).__contains__


class ShowesCollection(object):

    """A set of showes from a query."""

    #: The header of every column
    headers = '主题 票价 场馆 '.split()

    def __init__(self, rows):
        self._rows = rows

    def __repr__(self):
        return '<ShowTicketsCollection size={}>'.format(len(self))

    def __len__(self):
        return len(self._rows)

    def pretty_print(self):
        pt = PrettyTable()
        if len(self) == 0:
            pt._set_field_names(['Sorry,'])
            pt.add_row([SHOW_NOT_FOUND])
        else:
            pt._set_field_names(self.headers)
            for row in self._rows:
                pt.add_row(row)
        print(pt)


class ShowTicketsQuery(object):

    """Perform a show query."""

    def __init__(self, city, show_type, days=15):
        self.city = city
        self.show_type = show_type
        self.days = days

    def __repr__(self):
        return '<ShowTicketsQuery city={} types={} days={}>'.format(
            self._city, self._type, self._days
        )

    @property
    def cities(self):
        filepath = os.path.join(
            os.path.dirname(__file__),
            'datas', 'cities.dat'
        )
        d = {}
        with open(filepath, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                name, number = line.split()
                d.setdefault(name, int(number))
        return d

    @property
    def _city_id(self):
        ci = self.cities.get(self.city)
        if not ci:
            exit_after_echo(CITY_NOT_FOUND)
        return ci

    @property
    def _show_type(self):
        st = SHOW_TYPES.get(self.show_type)
        if not st:
            exit_after_echo(SHOW_NOT_FOUND)
        return st

    @property
    def date_range(self):
        """Generate date range according to the `days` user input."""
        try:
            days = int(self.days)
        except ValueError:
            exit_after_echo(QUERY_DAYS_INVALID)

        if days < 1:
            exit_after_echo(QUERY_DAYS_INVALID)
        start = datetime.today()
        end = start + timedelta(days=days)
        return (
            datetime.strftime(start, '%Y-%m-%d'),
            datetime.strftime(end, '%Y-%m-%d')
        )

    def _build_params(self):
        start, end = self.date_range
        return dict(
            cityID=self._city_id, isText=1,
            pageIndex=1, startDate=start,
            endDate=end, order=2,
            **self._show_type
        )

    def parse(self, items):
        """Parse `主题`, `时间`, `场馆`, 票价` in every item."""
        rows = []
        for i, item in enumerate(items):
            theme = colored.green(item.find(class_='ico').a.text.strip())
            text = item.find(class_='mt10').text.strip()
            mix = re.sub('\s+', ' ', text).split('：')
            time = mix[1][:-3]
            place = mix[2][:-7]
            # display time below theme
            theme_time = '\n'.join([
                theme,
                colored.red(''.join([
                    '(', time, ')'
                ]))
            ])
            price = item.find(class_='price-sort').text.strip()
            rows.append([theme_time, price, place])
        return rows

    def query(self):
        params = self._build_params()
        rows = []

        while True:
            try:
                r = requests.get(SHOWES_QUERY_URL, params=params)
            except ConnectionError:
                exit_after_echo(NETWORK_CONNECTION_FAIL)

            soup = BeautifulSoup(r.text, 'html.parser')
            items = soup.find_all(class_='ri-infos')

            if not items:
                return ShowesCollection(rows)

            rows += self.parse(items)
            params['pageIndex'] += 1


def query(params):
    """`params` is a list, contains `city`, 'show_type`, `days`."""

    return ShowTicketsQuery(*params).query()



# -*- coding: utf-8 -*-

"""
iquery.trains
~~~~~~~~~~~~~~

Train tickets query and display. The datas come
from:
    www.12306.cn
"""

import os
import re
import sys
import json
import requests
from .utils import colored, exit_after_echo
from datetime import datetime
from collections import OrderedDict
from prettytable import PrettyTable
from requests.exceptions import ConnectionError


__all__ = ['query']


QUERY_URL = 'https://kyfw.12306.cn/otn/lcxxcx/query'

# ERROR MSG
FROM_STATION_NOT_FOUND = 'From station not found.'
TO_STATION_NOT_FOUND = 'To station not found.'
INVALID_DATE = 'Invalid query date.'
NETWORK_CONNECTION_FAIL = 'Network connection failed.'
TRAIN_NOT_FOUND = 'No result.'
NO_RESPONSE = 'Sorry, server is not responding.'


class TrainsCollection(object):

    """A set of raw datas from a query."""

    headers = '车次 车站 时间 历时 商务 一等 二等 软卧 硬卧 软座 硬座 无座'.split()

    def __init__(self, rows, opts):
        self._rows = rows
        self._opts = opts

    def __repr__(self):
        return '<TrainsCollection size={}>'.format(len(self))

    def __iter__(self):
        i = 0
        while True:
            if i < len(self):
                yield self[i]
            else:
                yield next(self)
            i += 1

    def __len__(self):
        return len(self._rows)

    def _get_duration(self, row):
        duration = row.get('lishi').replace(':', '小时') + '分钟'
        # take 0 hour , only show minites
        if duration.startswith('00'):
            return duration[4:]
        # take <10 hours, show 1 bit
        if duration.startswith('0'):
            return duration[1:]
        return duration

    @property
    def trains(self):
        """Filter rows according to `headers`"""
        for row in self._rows:
            train_no = row.get('station_train_code')
            initial = train_no[0].lower()
            if not self._opts or initial in self._opts:
                train = [
                    # Column: '车次'
                    train_no,
                    # Column: '车站'
                    '\n'.join([
                        colored.green(row.get('from_station_name')),
                        colored.red(row.get('to_station_name')),
                    ]),
                    # Column: '时间'
                    '\n'.join([
                        colored.green(row.get('start_time')),
                        colored.red(row.get('arrive_time')),
                    ]),
                    # Column: '历时'
                    self._get_duration(row),
                    # Column: '商务'
                    row.get('swz_num'),
                    # Column: '一等'
                    row.get('zy_num'),
                    # Column: '二等'
                    row.get('ze_num'),
                    # Column: '软卧'
                    row.get('rw_num'),
                    # Column: '硬卧'
                    row.get('yw_num'),
                    # Column: '软座'
                    row.get('rz_num'),
                    # Column: '硬座'
                    row.get('yz_num'),
                    # Column: '无座'
                    row.get('wz_num')
                ]
                yield train

    def pretty_print(self):
        """Use `PrettyTable` to perform formatted outprint."""
        pt = PrettyTable()
        if len(self) == 0:
            pt._set_field_names(['Sorry,'])
            pt.add_row([TRAIN_NOT_FOUND])
        else:
            pt._set_field_names(self.headers)
            for train in self.trains:
                pt.add_row(train)
        print(pt)


class TrainTicketsQuery(object):

    """Docstring for TrainTicketsCollection. """

    def __init__(self, from_station, to_station, date, opts=None):

        self.from_station = from_station
        self.to_station = to_station
        self.date = date
        self.opts = opts

    def __repr__(self):
        return 'TrainTicketsQuery from={} to={} date={}'.format(
            self._from_station, self._to_station, self._date
        )

    @property
    def stations(self):
        filepath = os.path.join(
            os.path.dirname(__file__),
            'datas', 'stations.dat'
        )
        d = {}
        with open(filepath, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                name, telecode = line.split()
                d.setdefault(name, telecode)
        return d

    @property
    def _from_station_telecode(self):
        code = self.stations.get(self.from_station)
        if not code:
            exit_after_echo(FROM_STATION_NOT_FOUND)
        return code

    @property
    def _to_station_telecode(self):
        code = self.stations.get(self.to_station)
        if not code:
            exit_after_echo(TO_STATION_NOT_FOUND)
        return code

    @property
    def _valid_date(self):
        """Check and return a valid query date."""
        date = self._parse_date(self.date)

        if not date:
            exit_after_echo(INVALID_DATE)

        try:
            date = datetime.strptime(date, '%Y%m%d')
        except ValueError:
            exit_after_echo(INVALID_DATE)

        # A valid query date should within 50 days.
        offset = date - datetime.today()
        if offset.days not in range(-1, 50):
            exit_after_echo(INVALID_DATE)

        return datetime.strftime(date, '%Y-%m-%d')

    @staticmethod
    def _parse_date(date):
        """Parse from the user input `date`.

        e.g. current year 2016:
           input 6-26, 626, ... return 2016626
           input 2016-6-26, 2016/6/26, ... retrun 2016626

        This fn wouldn't check the date, it only gather the number as a string.
        """
        result = ''.join(re.findall('\d', date))
        l = len(result)

        # User only input month and day, eg 6-1, 6.26, 0626...
        if l in (2, 3, 4):
            year = str(datetime.today().year)
            return year + result

        # User input full format date, eg 201661, 2016-6-26, 20160626...
        if l in (6, 7, 8):
            return result

        return ''

    def _build_params(self):
        """Have no idea why wrong params order can't get data.
        So, use `OrderedDict` here.
        """
        d = OrderedDict()
        d['purpose_codes'] = 'ADULT'
        d['queryDate'] = self._valid_date
        d['from_station'] = self._from_station_telecode
        d['to_station'] = self._to_station_telecode
        return d

    def query(self):

        params = self._build_params()

        try:
            r = requests.get(QUERY_URL, params=params, verify=False)
        except ConnectionError:
            exit_after_echo(NETWORK_CONNECTION_FAIL)

        try:
            rows = r.json()['data']['datas']
        except KeyError:
            rows = []
        except TypeError:
            exit_after_echo(NO_RESPONSE)

        return TrainsCollection(rows, self.opts)


def query(params):
    """`params` is a list, contains `from`, `to`, `date`."""

    return TrainTicketsQuery(*params).query()



# -*- coding: utf-8 -*-

"""
iquery.putian_hospitals
~~~~~~~~~~~~~~~~~~~~~~

Check whether a hospital is putian series or display
all putian series hospitals in a city. The datas come
from:
    https://github.com/open-power-workgroup/Hospital
"""

import requests
import itertools
from .utils import args, colored, exit_after_echo
from prettytable import PrettyTable
from requests.exceptions import ConnectionError, Timeout


__all__ = ['query']


QUERY_URL = ('https://raw.githubusercontent.com/'
             'open-power-workgroup/Hospital/master/'
             'resource/API_resource/hospital_list.json')


class HospitalCollection(object):

    """Putian hospitals data sets."""

    def __init__(self, rows, params):
        self._rows = rows

        self._params = params

        #: All supported cities.
        self._citys = rows.keys()

    def pretty_print(self):

        pt = PrettyTable()

        l = len(self._params)

        city = self._params[0]
        if city not in self._citys:
            exit_after_echo('City is not supported.')
        hospitals = self._rows.get(city)

        if l == 1:
            pt._set_field_names([city])
            for idx, h in enumerate(hospitals):
                color = colored.red
                pt.add_row([color(h) + '\n'])
            print(pt)
            return

        if l == 2:
            hospital = self._params[1]
            field_name = city + hospital
            is_putian = False
            for h in hospitals:
                if hospital in h:
                    is_putian = True
                    field_name = h
                    break
            pt._set_field_names([field_name])
            color = colored.red if is_putian else colored.green
            pt.add_row([color(str(is_putian))])
            print(pt)
            return


def query(params):
    """`params` is a `hospital` name  or a `city` name."""

    try:
        r = requests.get(QUERY_URL)
    except ConnectionError:
        exit_after_echo('Network connection failed.')

    return HospitalCollection(r.json(), params)



# -*- coding: utf-8 -*-

"""
iquery.core
~~~~~~~~~~~~

The program entrance.
"""

from .utils import args, exit_after_echo


try:
    from requests.packages.urllib3.exceptions import (
        SNIMissingWarning,
        InsecureRequestWarning,
        InsecurePlatformWarning
    )

# Not show warings
    requests.packages.urllib3.disable_warnings(SNIMissingWarning)
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
    requests.packages.urllib3.disable_warnings(InsecurePlatformWarning)
except ImportError:
    pass


def show_usage():
    """Usage:
    iquery (-m|电影)
    iquery -p <city>
    iquery -p <city> <hospital>
    iquery <city> <show> [<days>]
    iquery [-dgktz] <from> <to> <date>

Go to `tickets -h` for more details.
"""
    pass


def cli():
    """Various information query via command line.

Usage:
    iquery (-m|电影)
    iquery -p <city>
    iquery -p <city> <hospital>
    iquery <city> <show> [<days>]
    iquery [-dgktz] <from> <to> <date>

Arguments:
    from             出发站
    to               到达站
    date             查询日期

    city             查询城市
    show             演出的类型
    days             查询近(几)天内的演出, 若省略, 默认15

    city             城市名,加在-p后查询该城市所有莆田医院
    hospital         医院名,加在city后检查该医院是否是莆田系


Options:
    -h, --help       显示该帮助菜单.

    -dgktz           动车,高铁,快速,特快,直达

    -m               热映电影查询
    -p               莆田系医院查询

Show:
    演唱会 音乐会 音乐剧 歌舞剧 儿童剧 话剧
    歌剧 比赛 舞蹈 戏曲 相声 杂技 马戏 魔术


Go to https://github.com/protream/tickets for usage examples.
"""

    if args.is_asking_for_help:
        exit_after_echo(cli.__doc__, color=None)

    elif args.is_querying_movie:
        from .movies import query
        result = query()

    elif args.is_querying_show:
        from .showes import query
        result = query(args.as_show_query_params)

    elif args.is_querying_putian_hospital:
        from .hospitals import query
        result = query(args.as_hospital_query_params)

    elif args.is_querying_train:
        from .trains import query
        result = query(args.as_train_query_params)

    else:
        exit_after_echo(show_usage.__doc__, color=None)

    result.pretty_print()



# -*- coding: utf-8 -*-

"""
iquery.movies
~~~~~~~~~~~~~~

Movies query and display. The datas come
from:
    m.douban.com/movie
"""

import re
import textwrap
import requests
from .utils import colored, exit_after_echo
from bs4 import BeautifulSoup
from prettytable import PrettyTable
from requests.exceptions import ConnectionError


__all__ = ['query']


QUERY_URL = ('https://frodo.douban.com/jsonp/'
             'subject_collection/movie_showing/items')

NETWORK_CONNECTION_FAIL = 'Network connection failed.'


class MoviesCollection(object):

    """Docstring for HotAndComingMovies. """

    header = '编号 电影名称+上映日期 导演+主演+类型 豆瓣评分'.split()

    def __init__(self, rows):
        self._rows = rows

    def __len__(self):
        return len(self._rows)

    @property
    def movies(self):
        for idx, row in enumerate(self._rows):
            rating = row.get('rating')
            # douban score
            score = '{:.1f}'.format(rating.get('value')) if rating else '暂无'
            infos = row['info'].split('/')
            if re.match('\d', infos[-1]):
                time = infos[-1:]
                infos = '/'.join(infos[:-1])
            else:
                time = infos[-2:]
                infos = '/'.join(infos[:-2])
            m = [
                idx + 1,
                '\n'.join([
                    colored.green(row['title']),
                    colored.red(''.join([
                        '(', time[0][:10], ')'
                    ])),
                ]),
                infos,
                score
            ]
            yield m

    def _get_movie_summary(self, num):
        url = self._rows[num - 1].get('url')
        try:
            r = requests.get(url)
        except ConnectionError:
            exit_after_echo(NETWORK_CONNECTION_FAIL)
        soup = BeautifulSoup(r.text, 'html.parser')
        s = re.sub(r'\s+', '', soup.find(property="v:summary").text)
        print(textwrap.fill(colored.green(s), 40, initial_indent=''))

    def pretty_print(self):

        pt = PrettyTable()
        pt._set_field_names(self.header)
        for m in self.movies:
            pt.add_row(m)
        print(pt)

        print('输入编号获取剧情简介:')
        while True:
            raw = input('>> ')
            if raw in ('q', 'quit'):
                exit()
            try:
                num = int(raw)
            except ValueError:
                print('Invalid number.')
                continue

            if (num - 1) in range(len(self)):
                self._get_movie_summary(num)
            else:
                print('Invalid number.')


def query():
    """Query hot movies infomation from douban."""

    try:
        r = requests.get(QUERY_URL)
    except ConnectionError:
        exit_after_echo(NETWORK_CONNECTION_FAIL)

    try:
        rows = r.json()['subject_collection_items']
    except (IndexError, TypeError):
        rows = []

    return MoviesCollection(rows)
